#include "constraint.h"
#include "chip_math.h"
#include "impl.h"

Constraint::Constraint()
{
	CHIPPER_INIT;
	cpcon = 0;
	bodies[0] = bodies[1] = 0;
}

Constraint::~Constraint()
{
	if(cpcon) {
		cpConstraintFree(cpcon);
	}
}

cpConstraint *Constraint::chipmunk_constraint() const
{
	return cpcon;
}

void Constraint::set_bodies(Body *b1, Body *b2)
{
	bodies[0] = b1;
	bodies[1] = b2;
	cpcon->a = b1->chipmunk_body();
	cpcon->b = b2->chipmunk_body();
}

Body *Constraint::get_body(int idx) const
{
	if(idx < 0 || idx > 2) {
		return 0;
	}
	return bodies[idx];
}

void Constraint::set_max_force(float max_force)
{
	cpcon->maxForce = max_force;
}

float Constraint::get_max_force() const
{
	return cpcon->maxForce;
}

void Constraint::set_bias_coef(float bias_coef)
{
	cpcon->biasCoef = bias_coef;
}

float Constraint::get_bias_coef() const
{
	return cpcon->biasCoef;
}

// ----- pivot joint -----
PivotJoint::PivotJoint(const Vector2 &pivot)
{
	cpcon = cpPivotJointNew(0, 0, chip_vec2(pivot));
}

PivotJoint::PivotJoint(const Vector2 &p1, const Vector2 &p2)
{
	cpcon = cpPivotJointNew2(0, 0, chip_vec2(p1), chip_vec2(p2));
}

PivotJoint::PivotJoint(Body *b1, Body *b2, const Vector2 &pivot)
{
	cpcon = cpPivotJointNew(b1->chipmunk_body(), b2->chipmunk_body(), chip_vec2(pivot));
	set_bodies(b1, b2);
}

PivotJoint::PivotJoint(Body *b1, Body *b2, const Vector2 &p1, const Vector2 &p2)
{
	cpcon = cpPivotJointNew2(b1->chipmunk_body(), b2->chipmunk_body(), chip_vec2(p1),
			chip_vec2(p2));
	set_bodies(b1, b2);
}

PivotJoint::~PivotJoint() {}

void PivotJoint::set_pivot(const Vector2 &pivot)
{
	((cpPivotJoint*)cpcon)->anchr1 = ((cpPivotJoint*)cpcon)->anchr2 = chip_vec2(pivot);
}

void PivotJoint::set_pivot(const Vector2 &p1, const Vector2 &p2)
{
	((cpPivotJoint*)cpcon)->anchr1 = chip_vec2(p1);
	((cpPivotJoint*)cpcon)->anchr2 = chip_vec2(p2);
}

Vector2 PivotJoint::get_pivot(int idx) const
{
	if(!idx) {
		return vmath_vec2(((cpPivotJoint*)cpcon)->anchr1);
	}
	return vmath_vec2(((cpPivotJoint*)cpcon)->anchr2);
}

// ---- rotational limit joint -----

RotLimit::RotLimit(float min_rot, float max_rot)
{
	cpcon = cpRotaryLimitJointNew(0, 0, min_rot, max_rot);
}

RotLimit::RotLimit(Body *b1, Body *b2, float min_rot, float max_rot)
{
	cpcon = cpRotaryLimitJointNew(b1->chipmunk_body(), b2->chipmunk_body(), min_rot, max_rot);
	set_bodies(b1, b2);
}

RotLimit::~RotLimit() {}

void RotLimit::set_limits(float min_rot, float max_rot)
{
	((cpRotaryLimitJoint*)cpcon)->min = min_rot;
	((cpRotaryLimitJoint*)cpcon)->max = max_rot;
}

float RotLimit::get_limit_min() const
{
	return ((cpRotaryLimitJoint*)cpcon)->min;
}

float RotLimit::get_limit_high() const
{
	return ((cpRotaryLimitJoint*)cpcon)->max;
}

// ---- spring constraint -----

Spring::Spring(float rlen, float k, float damping)
{
	cpcon = cpDampedSpringNew(0, 0, cpv(0, 0), cpv(0, 0), rlen, k, damping);
}

Spring::Spring(Body *b1, Body *b2, float rlen, float k, float damping)
{
	cpcon = cpDampedSpringNew(0, 0, cpv(0, 0), cpv(0, 0), rlen, k, damping);
	set_bodies(b1, b2);
}

Spring::~Spring() {}

void Spring::set_anchors(const Vector2 &a1, const Vector2 &a2)
{
	((cpDampedSpring*)cpcon)->anchr1 = chip_vec2(a1);
	((cpDampedSpring*)cpcon)->anchr2 = chip_vec2(a2);
}

void Spring::set_anchor(int idx, const Vector2 &a)
{
	if(idx < 0 || idx > 1) {
		return;
	}
	if(idx == 0) {
		((cpDampedSpring*)cpcon)->anchr1 = chip_vec2(a);
	} else {
		((cpDampedSpring*)cpcon)->anchr2 = chip_vec2(a);
	}
}

Vector2 Spring::get_anchor(int idx)
{
	if(idx == 0) {
		return vmath_vec2(((cpDampedSpring*)cpcon)->anchr1);
	} else if(idx == 1) {
		return vmath_vec2(((cpDampedSpring*)cpcon)->anchr2);
	}
	return Vector2(0, 0);	// error
}

void Spring::set_stiffness(float k)
{
	((cpDampedSpring*)cpcon)->stiffness = k;
}

float Spring::get_stiffness() const
{
	return ((cpDampedSpring*)cpcon)->stiffness;
}

void Spring::set_rest_length(float rlen)
{
	((cpDampedSpring*)cpcon)->restLength = rlen;
}

float Spring::get_rest_length() const
{
	return ((cpDampedSpring*)cpcon)->restLength;
}

void Spring::set_damping(float d)
{
	((cpDampedSpring*)cpcon)->damping = d;
}

float Spring::get_damping() const
{
	return ((cpDampedSpring*)cpcon)->damping;
}