#include "shape.h"
#include "impl.h"
#include "chip_math.h"
#include "chipper.h"

Shape::Shape(Body *body)
{
	CHIPPER_INIT;

	cpshape = 0;
	this->body = body;
}

Shape::~Shape() {}

cpShape *Shape::chipmunk_shape() const
{
	return cpshape;
}

void Shape::set_collision_type(unsigned int type)
{
	cpshape->collision_type = type;
}

unsigned int Shape::get_collision_type() const
{
	return cpshape->collision_type;
}

void Shape::set_group(unsigned int group)
{
	cpshape->group = group;
}

unsigned int Shape::get_group() const
{
	return cpshape->group;
}

void Shape::set_layer_mask(unsigned int mask)
{
	cpshape->layers = mask;
}

unsigned int Shape::get_layer_mask() const
{
	return cpshape->layers;
}

void Shape::add_to_layer(int layer)
{
	cpshape->layers |= (1 << layer);
}

void Shape::remove_from_layer(int layer)
{
	cpshape->layers &= ~(1 << layer);
}

bool Shape::in_layer(int layer) const
{
	return (cpshape->layers & (1 << layer)) != 0;
}

void Shape::set_elasticity(float e)
{
	cpshape->e = e;
}

float Shape::get_elasticity() const
{
	return cpshape->e;
}

void Shape::set_friction(float f)
{
	cpshape->u = f;
}

float Shape::get_friction() const
{
	return cpshape->u;
}

void Shape::add_to_body(Body *body)
{
	this->body = body;
	cpshape->body = body->chipmunk_body();
}

void Shape::remove_from_body()
{
	body = 0;
	cpshape->body = 0;
}

Body *Shape::get_body() const
{
	return body;
}

bool Shape::contains_point(const Vector2 &pt) const
{
	return cpShapePointQuery(cpshape, chip_vec2(pt)) != 0;
}

// ---- circle ----

Circle::Circle(Body *body, float rad, const Vector2 &center)
	: Shape(body)
{
	cpshape = cpCircleShapeNew(body->chipmunk_body(), rad, chip_vec2(center));
}

Circle::~Circle()
{
	cpShapeFree(cpshape);
}

void Circle::set_radius(float rad)
{
	((cpCircleShape*)cpshape)->r = rad;
}

float Circle::get_radius() const
{
	return ((cpCircleShape*)cpshape)->r;
}

void Circle::set_center(const Vector2 &center)
{
	((cpCircleShape*)cpshape)->c = chip_vec2(center);
}

Vector2 Circle::get_center() const
{
	return vmath_vec2(((cpCircleShape*)cpshape)->c);
}

float Circle::calc_moment() const
{
	if(!body) {
		return 0.0;
	}
	return calc_moment_circle(body->get_mass(), get_radius());
}

// ------ line segments ------

Segment::Segment(Body *body, const Vector2 &v1, const Vector2 &v2, float rad)
	: Shape(body)
{
	cpshape = cpSegmentShapeNew(body->chipmunk_body(), chip_vec2(v1), chip_vec2(v2), rad);
}

Segment::~Segment() {}

void Segment::set_endpoints(const Vector2 &v1, const Vector2 &v2)
{
	((cpSegmentShape*)cpshape)->a = chip_vec2(v1);
	((cpSegmentShape*)cpshape)->b = chip_vec2(v2);
}

Vector2 Segment::get_endpoint(int idx) const
{
	switch(idx) {
	case 0:
		return vmath_vec2(((cpSegmentShape*)cpshape)->a);
	case 1:
		return vmath_vec2(((cpSegmentShape*)cpshape)->b);
	default:
		break;
	}
	return Vector2(0, 0);	// error
}

void Segment::set_radius(float r)
{
	((cpSegmentShape*)cpshape)->r = r;
}

float Segment::get_radius() const
{
	return ((cpSegmentShape*)cpshape)->r;
}

float Segment::calc_moment() const
{
	if(!body) {
		return 0.0;
	}
	return calc_moment_segment(body->get_mass(), get_endpoint(0), get_endpoint(1));
}

// ----- polygon ----
void setUpVerts(cpPolyShape *poly, int numVerts, cpVect *verts, cpVect offset);

Poly::Poly(Body *body, const Vector2 *varr, int num_verts)
	: Shape(body)
{
	cpVect *v = new cpVect[num_verts];
	for(int i=0; i<num_verts; i++) {
		v[i].x = varr[i].x;
		v[i].y = varr[i].y;
	}

	cpshape = cpPolyShapeNew(body->chipmunk_body(), num_verts, v, cpv(0, 0));
	delete [] v;
}

Poly::~Poly() {}

void Poly::set_vertex_data(const Vector2 *varr, int num_verts)
{
	cpVect *v = new cpVect[num_verts];
	for(int i=0; i<num_verts; i++) {
		v[i].x = varr[i].x;
		v[i].y = varr[i].y;
	}

	setUpVerts((cpPolyShape*)cpshape, num_verts, v, cpv(0, 0));
	delete [] v;
}

int Poly::get_vertex_count() const
{
	return ((cpPolyShape*)cpshape)->numVerts;
}

float Poly::calc_moment() const
{
	if(!body) {
		return 0.0;
	}
	return calc_moment_polygon(body->get_mass(), (Vector2*)((cpPolyShape*)cpshape)->verts, get_vertex_count());
}
