#include <algorithm>
#include "simspace.h"
#include "impl.h"
#include "chip_math.h"

using namespace std;

bool chipper_was_initialized;

SimSpace::SimSpace()
{
	CHIPPER_INIT;

	space = cpSpaceNew();

	upd_interval = 0.016;
	prev_upd = 0;
}

SimSpace::~SimSpace()
{
	for(size_t i=0; i<bodies.size(); i++) {
		delete bodies[i];
	}
	for(size_t i=0; i<shapes.size(); i++) {
		delete shapes[i];
	}
	for(size_t i=0; i<joints.size(); i++) {
		delete joints[i];
	}
	cpSpaceFree(space);
}

void SimSpace::set_gravity(const Vector2 &g)
{
	space->gravity = chip_vec2(g);
}

Vector2 SimSpace::get_gravity() const
{
	return vmath_vec2(space->gravity);
}

void SimSpace::set_iterations(int iter)
{
	space->iterations = iter;
}

int SimSpace::get_iterations() const
{
	return space->iterations;
}

void SimSpace::obj_size_hint(float sz)
{
	cpSpaceResizeStaticHash(space, sz, 1000);
	cpSpaceResizeActiveHash(space, sz, 1000);
}

void SimSpace::add_body(Body *body)
{
	bodies.push_back(body);
	cpSpaceAddBody(space, body->chipmunk_body());
}

void SimSpace::add_shape(Shape *shape)
{
	shapes.push_back(shape);
	cpSpaceAddShape(space, shape->chipmunk_shape());

	cpshape_map[shape->chipmunk_shape()] = shape;
}

void SimSpace::add_static_shape(Shape *shape)
{
	shapes.push_back(shape);
	cpSpaceAddStaticShape(space, shape->chipmunk_shape());
}

void SimSpace::add_constraint(Constraint *constr)
{
	joints.push_back(constr);
	cpSpaceAddConstraint(space, constr->chipmunk_constraint());
}

bool SimSpace::remove_body(Body *body)
{
	vector<Body*>::iterator iter = find(bodies.begin(), bodies.end(), body);
	if(iter == bodies.end()) {
		return false;
	}
	bodies.erase(iter);
	cpSpaceRemoveBody(space, body->chipmunk_body());
	return true;
}

bool SimSpace::remove_shape(Shape *shape)
{
	vector<Shape*>::iterator iter = find(shapes.begin(), shapes.end(), shape);
	if(iter == shapes.end()) {
		return false;
	}
	shapes.erase(iter);
	cpSpaceRemoveShape(space, shape->chipmunk_shape());

	map<cpShape*, Shape*>::iterator miter = cpshape_map.find(shape->chipmunk_shape());
	if(miter != cpshape_map.end()) {
		cpshape_map.erase(miter);
	}
	return true;
}

bool SimSpace::remove_static_shape(Shape *shape)
{
	vector<Shape*>::iterator iter = find(shapes.begin(), shapes.end(), shape);
	if(iter == shapes.end()) {
		return false;
	}
	shapes.erase(iter);
	cpSpaceRemoveStaticShape(space, shape->chipmunk_shape());
	return true;
}

bool SimSpace::remove_constraint(Constraint *constr)
{
	vector<Constraint*>::iterator iter = find(joints.begin(), joints.end(), constr);
	if(iter == joints.end()) {
		return false;
	}
	joints.erase(iter);
	cpSpaceRemoveConstraint(space, constr->chipmunk_constraint());
	return true;
}

void SimSpace::clear_bodies(void (*destr)(Body*))
{
	for(size_t i=0; i<bodies.size(); i++) {
		cpSpaceRemoveBody(space, bodies[i]->chipmunk_body());
		if(destr) {
			destr(bodies[i]);
		}
	}
	bodies.clear();
}

void SimSpace::clear_shapes(void (*destr)(Shape*))
{
	vector<Shape*>::iterator iter = shapes.begin();
	while(iter != shapes.end()) {
		Shape *s = *iter;

		map<cpShape*, Shape*>::iterator mapiter = cpshape_map.find(s->chipmunk_shape());
		if(mapiter != cpshape_map.end()) {
			cpSpaceRemoveShape(space, s->chipmunk_shape());
			if(destr) {
				destr(s);
			}
			iter = shapes.erase(iter);
			cpshape_map.erase(mapiter);
		} else {
			iter++;
		}
	}
}

void SimSpace::clear_static_shapes(void (*destr)(Shape*))
{
	vector<Shape*>::iterator iter = shapes.begin();
	while(iter != shapes.end()) {
		Shape *s = *iter;

		if(cpshape_map.find(s->chipmunk_shape()) == cpshape_map.end()) {
			cpSpaceRemoveStaticShape(space, s->chipmunk_shape());
			if(destr) {
				destr(s);
			}
			iter = shapes.erase(iter);
		} else {
			iter++;
		}
	}
}

void SimSpace::clear_constraints(void (*destr)(Constraint*))
{
	for(size_t i=0; i<joints.size(); i++) {
		cpSpaceRemoveConstraint(space, joints[i]->chipmunk_constraint());
		if(destr) {
			destr(joints[i]);
		}
	}
	joints.clear();
}

void SimSpace::set_update_interval(float sec)
{
	upd_interval = sec;
}

void SimSpace::update(float sec)
{
	float interval = sec - prev_upd;

	if(interval < upd_interval) {
		return;
	}

	while(interval >= upd_interval) {
		cpSpaceStep(space, upd_interval);
		interval -= upd_interval;
	}

	prev_upd = sec + interval;
}

struct QueryData {
	vector<Shape*> *svec;
	const map<cpShape*, Shape*> *smap;
};

static void query_func(cpShape *cpshape, void *data)
{
	QueryData *qdat = (QueryData*)data;

	map<cpShape*, Shape*>::const_iterator iter = qdat->smap->find(cpshape);
	if(iter != qdat->smap->end()) {
		qdat->svec->push_back(iter->second);
	}
}

Shape **SimSpace::find_shapes(const Vector2 &pos, int *count) const
{
	QueryData qdat = {&fres, &cpshape_map};

	fres.clear();
	cpSpaceShapePointQuery(space, chip_vec2(pos), query_func, (void*)&qdat);

	if(fres.empty()) {
		return 0;
	}
	if(count) {
		*count = fres.size();
	}
	return &fres[0];
}
