#include <vector>
#include "opengl.h"
#include "parts.h"
#include "polyline.h"
#include "chalk.h"
#include "chipper.h"
#include "ragdoll.h"
#include "palette.h"
#include "text.h"
#include "level.h"
#include "message.h"
#include "options.h"
#include "filedialog.h"
#include "image.h"

#define SCRSZ		1
#define NORM_X(x)	(SCRSZ * (aspect * (2.0 * (float)(x) / (float)width - 1.0)))
#define NORM_Y(y)	(SCRSZ * (1.0 - 2.0 * (float)(y) / (float)height))


static void disp();
static void draw_text();
static void draw_help();
static void idle();
static void keyb(unsigned char key, int x, int y);
static void skeyb(int key, int x, int y);
static void mouse(int bn, int state, int x, int y);
static void motion(int x, int y);
static void capture_thumb();

extern void draw_backdrop(float pan);
extern Vector2 mouse_to_space(int x, int y);

static float pan;

static Level level;

static PolyLine *cur_line;
static float cur_line_width = 0.04;

static int cur_col;

static SimSpace *space;

static Vector2 mouse_point, mouse_point_last;
static Body *mouse_body;
static Constraint *mouse_joint;

static bool inp_drawing = true;
static bool setting_start, setting_end;
static bool collidable_chalk = true;

extern int width, height;
extern float aspect;
extern unsigned int chalk_font;
extern Ragdoll *rag;

static bool have_ragdoll;
static bool resume_from_filedlg;

static bool show_help;

static unsigned int *thumb;
static int thumb_width, thumb_height;

static const char *help_text_lines[] = {
	"Controls",
	"----------",
	"draw by dragging the mouse",
	"pan by middle-button dragging",
	"right button toggles drawing / handling",
	"1-9 - pick chalk color",
	"tab - toggle collidable / non-collidable",
	"esc - exit without saving",
	"enter - save and exit",
	"S - save without exiting",
	"L - revert to last save",
	"C - clear level",
	"space - drop a ragdoll",
	"home/end - set start/goal",
	0
};

bool editor_init()
{
	// init sim
	space = new SimSpace;
	space->obj_size_hint(0.05);
	space->set_gravity(Vector2(0, -5));

	cp_collision_slop = 0.001;
	space->set_iterations(20);
	
	mouse_body = new Body(CHIPPER_INF, CHIPPER_INF);

	level.set_space(space);

	return true;
}

void editor_start()
{
	glutDisplayFunc(disp);
	glutKeyboardFunc(keyb);
	glutSpecialFunc(skeyb);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);
	glutIdleFunc(idle);
	glutSetCursor(GLUT_CURSOR_CROSSHAIR);

	rag->add_to_space(space);
	rag->set_color(get_color(get_player_color()));

	if(resume_from_filedlg) {
		const char *fname = fdlg_get_fname();
		FileDialogMode mode = fdlg_get_mode();

		if(fname) {
			if(mode == FDLG_LOAD) {
				if(!level.load(fname)) {
					show_message("failed to load level");
				} else {
					show_message("level loaded");
				}
			} else {
				if(!level.save(fname)) {
					show_message("failed to save level");
				} else {
					show_message("level saved");

					// also save a thumbnail
					if(thumb) {
						char *img_fname = new char[strlen(fname) + 5];
						sprintf(img_fname, "%s.png", fname);

						if(save_image(img_fname, thumb, thumb_width, thumb_height, IMG_FMT_PNG) == -1) {
							fprintf(stderr, "failed to save thumbnail\n");
						}

						delete [] img_fname;
						delete [] thumb;
						thumb = 0;
					}
				}
			}
		}
		resume_from_filedlg = false;
	} else {
		pan = 0.0;
	}
}

void editor_stop()
{
	rag->remove_from_space(space);

	glutKeyboardFunc(0);
	glutSpecialFunc(0);
	glutMouseFunc(0);
	glutMotionFunc(0);
	glutIdleFunc(0);
	glutSetCursor(GLUT_CURSOR_LEFT_ARROW);
}

static void disp()
{
	unsigned int msec = glutGet(GLUT_ELAPSED_TIME);
	glClear(GL_COLOR_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(-pan, 0, 0);

	draw_backdrop(pan);

	level.draw();

	if(have_ragdoll) {
		rag->draw();
		level.set_char_pos(rag->get_position());
	}

	draw_text();
	if(show_help) {
		draw_help();
	}

	// --- ragdoll manipulation ---
	/*if(!inp_drawing) {
		Vector2 new_point = mouse_point_last + (mouse_point - mouse_point_last) * 0.25;
		mouse_body->set_position(new_point);
		mouse_body->set_velocity((new_point - mouse_point_last) * 60.0);
		mouse_point_last = new_point;
	}*/

	draw_msg();

	// --- simulation stepping ---
	//space->update(msec / 1000.0);

	glutSwapBuffers();
}

#define TEXT_Y	0.035
static void draw_text()
{
	bind_font(chalk_font);
	set_text_color(1, 1, 1, 0.75);
	set_text_pos(0.01, TEXT_Y);
	set_text_size(0.75);
	//print_string("editing level, press h for help");

	set_text_pos(0.5, TEXT_Y);

	Color col = get_color(cur_col);
	set_text_color(col.x, col.y, col.z, 0.75);
	print_string(get_color_name(cur_col));
	print_string(" chalk");

	set_text_pos(0.75, TEXT_Y);

	set_text_color(1, 1, 1, 0.75);
	print_string(collidable_chalk ? "collidable" : "non-collidable");

	set_text_size(1.0);
}

static void draw_help()
{
	glPushAttrib(GL_ENABLE_BIT);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glScalef(aspect, 1, 1);

	glBegin(GL_QUADS);
	glColor4f(0, 0, 0, 0.5);
	glVertex2f(-1, -1);
	glVertex2f(1, -1);
	glVertex2f(1, 1);
	glVertex2f(-1, 1);
	glEnd();

	glPopMatrix();
	glPopAttrib();

	set_text_color(1, 1, 1, 0.75);
	set_text_size(1.0);

	float ypos = 0.15;
	for(int i=0; help_text_lines[i]; i++) {
		set_text_pos(0.2, ypos);
		print_string(help_text_lines[i]);
		ypos += get_line_advance() * 1.5;
	}
}

static void idle()
{
	glutPostRedisplay();
}

static void keyb(unsigned char key, int x, int y)
{
	switch(key) {
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
		cur_col = key - '1';
		break;

	case '`':
		if(++cur_col > 7) {
			cur_col = 0;
		}
		break;

	case ' ':
		have_ragdoll = true;
		rag->reset();
		rag->set_position(level.get_start() + Vector2(0.0, 1.0));
		break;

	case '\t':
		collidable_chalk = !collidable_chalk;
		break;

	case 27:
		if(setting_start || setting_end) {
			glutSetCursor(inp_drawing ? GLUT_CURSOR_CROSSHAIR : GLUT_CURSOR_INFO);
			setting_start = setting_end = false;
		} else {
			switch_state(PART_MENU);
		}
		break;

	case 's':
	case 'S':
		capture_thumb();

		resume_from_filedlg = true;
		fdlg_set_mode(FDLG_SAVE);
		switch_state(PART_FDLG);
		break;

	case 'l':
	case 'L':
		resume_from_filedlg = true;
		fdlg_set_mode(FDLG_LOAD);
		switch_state(PART_FDLG);
		break;

	case 'h':
	case 'H':
		show_help = !show_help;
		break;

	case 'c':
	case 'C':
		level.reset();
		show_message("level cleared");
		break;
	}
}

static void skeyb(int key, int x, int y)
{
	switch(key) {
	case GLUT_KEY_HOME:
	case GLUT_KEY_F1:
		show_message("place start point");
		setting_start = true;
		glutSetCursor(GLUT_CURSOR_BOTTOM_SIDE);
		break;

	case GLUT_KEY_END:
	case GLUT_KEY_F2:
		show_message("place end point");
		setting_end = true;
		glutSetCursor(GLUT_CURSOR_BOTTOM_SIDE);
		break;
	}
}

static int pan_prev_x = -1;

static void mouse(int bn, int state, int x, int y)
{
	if(bn == GLUT_LEFT_BUTTON) {
		if(state == GLUT_DOWN) {
			if(setting_start || setting_end) {
				return;
			}
			if(inp_drawing) {
				// drawing mode
				if(!cur_line) {
					cur_line = new PolyLine;
					cur_line->set_color(get_color(cur_col));
					cur_line->set_width(cur_line_width);

					level.add_line(cur_line);
				}
				Vector2 pt(NORM_X(x), NORM_Y(y));
				cur_line->add_point(pt + Vector2(pan, 0));
			} else {
				// ragdoll manipulation mode
				Vector2 point = mouse_to_space(x, y);
				Shape **sres = space->find_shapes(point, 0);
				if(!sres) return;

				Body *body = sres[0]->get_body();

				mouse_joint = new PivotJoint(mouse_body, body, Vector2(0, 0), body->world_to_local(point));
				mouse_joint->set_max_force(5000.0);
				mouse_joint->set_bias_coef(0.15);
				space->add_constraint(mouse_joint);
			}
		} else {	// left button released...
			if(setting_start) {
				if(!level.set_start(Vector2(NORM_X(x) + pan, NORM_Y(y)))) {
					show_message("invalid placement, try again");
				} else {
					setting_start = false;
					glutSetCursor(inp_drawing ? GLUT_CURSOR_CROSSHAIR : GLUT_CURSOR_INFO);
				}
				return;
			}
			if(setting_end) {
				if(!level.set_end(Vector2(NORM_X(x) + pan, NORM_Y(y)))) {
					show_message("invalid placement, try again");
				} else {
					setting_end = false;
					glutSetCursor(inp_drawing ? GLUT_CURSOR_CROSSHAIR : GLUT_CURSOR_INFO);
				}
				return;
			}

			if(inp_drawing) {
				level.fin_line(cur_line, collidable_chalk, cur_col);
				cur_line = 0;
			} else {
				space->remove_constraint(mouse_joint);
				delete mouse_joint;
				mouse_joint = 0;
			}
		}
	}

	if(bn == GLUT_MIDDLE_BUTTON) {
		if(state == GLUT_DOWN) {
			pan_prev_x = x;
		} else {
			pan_prev_x = -1;
		}
	}

	/*if(bn == GLUT_RIGHT_BUTTON) {
		if(state == GLUT_DOWN) {
			inp_drawing = !inp_drawing;
			glutSetCursor(inp_drawing ? GLUT_CURSOR_CROSSHAIR : GLUT_CURSOR_INFO);
		}
	}*/
}

static void motion(int x, int y)
{
	if(inp_drawing && cur_line) {
		Vector2 pt(NORM_X(x), NORM_Y(y));

		// add the new point only if it's farther than 0.05 units away
		cur_line->add_point(pt + Vector2(pan, 0), 0.05);
	}

	if(pan_prev_x >= 0) {
		pan += (float)(pan_prev_x - x) / (float)(width / 2);
		pan_prev_x = x;
	}

	if(!inp_drawing) {
		mouse_point = mouse_to_space(x, y);
	}
}

static void capture_thumb()
{
	extern int width, height;

	unsigned int *shot = new unsigned int[width * height];
	glReadPixels(0, 0, width, height, GL_BGRA, GL_UNSIGNED_BYTE, shot);

	thumb_width = 256;
	thumb_height = 256;
	thumb = new unsigned int[thumb_width * thumb_height];

	float xscale = (float)width / (float)thumb_width;
	float yscale = (float)height / (float)thumb_height;

	unsigned int *dptr = thumb;
	for(int i=0; i<thumb_height; i++) {
		int src_y = (int)((float)(thumb_height - i - 1) * yscale);

		unsigned int *sptr = shot + src_y * width;

		for(int j=0; j<thumb_width; j++) {
			int src_x = (int)((float)j * xscale);

			*dptr++ = sptr[src_x];
		}
	}
	delete [] shot;
}
