#include <iostream>
#include "opengl.h"
#include "parts.h"
#include "level.h"
#include "message.h"
#include "text.h"
#include "options.h"
#include "ragdoll.h"
#include "palette.h"
#include "filedialog.h"

#define DEF_FNAME	"levels/level0.lvl"
#define HEIGHT		0.32
#define HEAD_OFFS	0.125

static void disp();
static void draw_text(unsigned int msec);
static void idle();
static void keyb(unsigned char key, int x, int y);
static void skeyb(int key, int x, int y);
static void skeyb_up(int key, int x, int y);
static void mouse(int bn, int state, int x, int y);
static void motion(int x, int y);
static void spawn(const Vector2 &where);
static void move(float dx);

void draw_backdrop(float pan);
Vector2 mouse_to_space(int, int);

static float pan;
static Level *level;
static unsigned int start_time;
static float prev_sec;

static SimSpace *space;
static Body *pin_body;
static Vector2 pin_point, pin_point_last;
static PivotJoint *pin_joint;

static bool left_key, right_key;
static bool ragdoll_moving;

static int exit_started = -1, exit_dur = 1000;

extern Ragdoll *rag;
extern unsigned int chalk_font;

bool game_init()
{
	pin_body = new Body(CHIPPER_INF, CHIPPER_INF);

	return true;
}

void game_start()
{
	glutDisplayFunc(disp);
	glutKeyboardFunc(keyb);
	glutSpecialFunc(skeyb);
	glutSpecialUpFunc(skeyb_up);
	//glutMouseFunc(mouse);
	//glutMotionFunc(motion);
	glutIdleFunc(idle);

	space = new SimSpace;
	space->obj_size_hint(0.05);
	space->set_gravity(Vector2(0, -3));

	cp_collision_slop = 0.001;
	space->set_iterations(20);
	space->set_update_interval(0.020);

	level = new Level;
	level->set_space(space);

	const char *fname = fdlg_get_fname();
	if(!fname) {
		fname = DEF_FNAME;
	}

	if(!level->load(fname)) {
		show_message("failed to load level!");
		return;
	}

	rag->add_to_space(space);
	rag->set_color(get_color(get_player_color()));

	spawn(level->get_start());

	//glutSetCursor(GLUT_CURSOR_INFO);
	glutSetCursor(GLUT_CURSOR_NONE);

	start_time = glutGet(GLUT_ELAPSED_TIME);
	show_message("Go!", 3000);
}

void game_stop()
{
	glutKeyboardFunc(0);
	glutSpecialFunc(0);
	glutMouseFunc(0);
	glutMotionFunc(0);
	glutIdleFunc(0);

	exit_started = -1;
	prev_sec = 0.0;

	rag->remove_from_space(space);

	delete level;
	delete space;

	glutSetCursor(GLUT_CURSOR_LEFT_ARROW);
}

static void disp()
{
	unsigned int msec = glutGet(GLUT_ELAPSED_TIME) - start_time;
	float sec = (float)msec / 1000.0;

	if(exit_started > 0 && msec - (unsigned int)exit_started > (unsigned int)exit_dur) {
		switch_state(PART_MENU);
		return;
	}

	float dx = (sec - prev_sec) * 0.25;
	prev_sec = sec;

	if(exit_started < 0 && left_key) {
		move(-dx);
	}
	if(exit_started < 0 && right_key) {
		move(dx);
	}

	Vector2 torso_pos = rag->get_body(BP_TORSO)->get_position();
	level->set_char_pos(torso_pos);
	//pan = torso_pos.x;

	if(exit_started < 0) {
		Vector2 dvec = level->get_end() - torso_pos;
		
		if(dvec.y > -0.5 && fabs(dvec.x) < 0.1) {
			exit_started = msec;
			exit_dur = 4000;
			show_message("well done!", exit_dur);
		}
	}

	// --- simulation stepping ---
	if(exit_started < 0) {
		space->update(sec);
	}

	glClear(GL_COLOR_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(-pan, 0, 0);

	draw_backdrop(pan);

	level->draw();
	rag->draw();

	draw_text(exit_started < 0 ? msec : exit_started);

	draw_msg();

	/*Vector2 new_point = pin_point_last + (pin_point - pin_point_last) * 0.25;
	pin_body->set_position(new_point);
	pin_body->set_velocity((new_point - pin_point_last) * 60.0);
	pin_point_last = new_point;*/

	pan = torso_pos.x;

	/*glPointSize(5.0);
	glBegin(GL_POINTS);
	glColor3f(1, 1, 0);
	glVertex2f(pin_point.x, pin_point.y);
	//glColor3f(0, 1, 1);
	//glVertex2f(new_point.x + pan, new_point.y);
	glEnd();*/

	glutSwapBuffers();
}

#define TEXT_Y	0.04
static void draw_text(unsigned int msec)
{
	const char *pname = get_player_name();

	bind_font(chalk_font);
	set_text_color(1, 1, 1, 0.75);
	set_text_pos(0.01, TEXT_Y);

	print_string(pname);

	unsigned int sec = msec / 1000;
	unsigned int min = sec / 60;
	msec -= sec * 1000;
	sec -= min * 60;
	float fsec = (float)sec + (float)(msec / 1000.0);

	char buf[64];
	sprintf(buf, "%dm %.1fs", min, fsec);

	set_text_pos(0.4, TEXT_Y);
	print_string(buf);
}

static void idle()
{
	glutPostRedisplay();
}

static void keyb(unsigned char key, int x, int y)
{
	switch(key) {
	case 27:
		switch_state(PART_MENU);
		break;
	}
}

static unsigned int start_fall;
static void skeyb(int key, int x, int y)
{
	if(glutGet(GLUT_ELAPSED_TIME) - start_time < start_fall + 1000) {
		return;
	}
	switch(key) {
	case GLUT_KEY_LEFT:
		left_key = true;
		rag->remove_from_space(space);
		ragdoll_moving = true;
		break;

	case GLUT_KEY_RIGHT:
		right_key = true;
		rag->remove_from_space(space);
		ragdoll_moving = true;
		break;
	}
}

static void skeyb_up(int key, int x, int y)
{
	switch(key) {
	case GLUT_KEY_LEFT:
		left_key = false;
		if(ragdoll_moving) {
			rag->add_to_space(space);
			ragdoll_moving = false;
		}
		break;

	case GLUT_KEY_RIGHT:
		right_key = false;
		if(ragdoll_moving) {
			rag->add_to_space(space);
			ragdoll_moving = false;
		}
		break;
	}
}

static void mouse(int bn, int state, int x, int y)
{
	if(exit_started > 0) {
		return;
	}
	if(bn == GLUT_LEFT_BUTTON) {
		if(state == GLUT_DOWN) {
			Vector2 point = mouse_to_space(x, y);
			Shape **sres = space->find_shapes(point, 0);
			if(!sres) return;

			Body *body = sres[0]->get_body();

			pin_joint = new PivotJoint(pin_body, body, Vector2(0, 0), body->world_to_local(point));
			pin_joint->set_max_force(5000.0);
			pin_joint->set_bias_coef(0.15);
			space->add_constraint(pin_joint);
		} else {
			space->remove_constraint(pin_joint);
			delete pin_joint;
			pin_joint = 0;
		}
	}
}

static void motion(int x, int y)
{
	if(exit_started > 0) {
		return;
	}
	pin_point = mouse_to_space(x, y);

	Vector2 gpt;
	if(level->get_ground(pin_point, &gpt, 0)) {
		if(pin_point.y - gpt.y > 0.4) {
			pin_point.y = gpt.y + 0.4;
		}
	}
}

static void spawn(const Vector2 &where)
{
	Vector2 pos = where;
	pos.y = 1.0;
	rag->set_position(pos);

	pan = pos.x;

	// set movement pin
	/*Body *head_body = rag->get_body(BP_HEAD);

	pin_point = mouse_to_space(width / 2.0, height / 2.0);

	pin_joint = new PivotJoint(pin_body, head_body, Vector2(0, 0),
			head_body->world_to_local(pin_point));
	pin_joint->set_max_force(5000.0);
	pin_joint->set_bias_coef(0.15);
	space->add_constraint(pin_joint);

	pan += pos.x;
	level->set_char_pos(pos);

	ragdoll_pinned = true;*/
}

static void move(float dx)
{
	Vector2 prev = rag->get_body(BP_TORSO)->get_position();

	Vector2 pos = prev + Vector2(dx, 0.0);
	Vector2 gnd;
	if(level->get_ground(pos, &gnd, 0)) {
		pos.y = gnd.y + HEIGHT;
	}

	if(prev.y - pos.y > 0.5) {
		ragdoll_moving = false;
		rag->add_to_space(space);
		start_fall = glutGet(GLUT_ELAPSED_TIME) - start_time;
	}

	rag->reset();
	rag->set_position(pos);
}
