#include <stdio.h>
#include <float.h>
#include "opengl.h"
#include "level.h"
#include "chalk.h"
#include "palette.h"

using namespace std;

static const float line_width = 0.05;

Level::Level()
{
	space = 0;
	static_body = new Body(CHIPPER_INF, CHIPPER_INF);

	start_set = end_set = false;
}

Level::~Level()
{
	reset();
}

void Level::reset()
{
	if(space) {
		for(size_t i=0; i<segm.size(); i++) {
			space->remove_static_shape(segm[i]);
			delete segm[i];
		}
		segm.clear();
	}

	for(size_t i=0; i<lines.size(); i++) {
		delete lines[i];
	}
	lines.clear();
	line_colid.clear();
	line_color.clear();

	start_set = end_set = false;
}

bool Level::save(const char *fname) const
{
	FILE *fp;

	if(!(fp = fopen(fname, "wb"))) {
		return false;
	}

	int num_lines = lines.size();
	fwrite(&num_lines, sizeof num_lines, 1, fp);

	for(int i=0; i<num_lines; i++) {
		int col = line_color[i];
		fwrite(&col, sizeof col, 1, fp);

		int collidable = line_colid[i] ? 1 : 0;
		fwrite(&collidable, sizeof collidable, 1, fp);

		int num_pt = lines[i]->get_point_count();
		fwrite(&num_pt, sizeof num_pt, 1, fp);

		for(int j=0; j<num_pt; j++) {
			Vector2 pt = lines[i]->get_point(j);

			fwrite(&pt.x, sizeof pt.x, 1, fp);
			fwrite(&pt.y, sizeof pt.y, 1, fp);
		}
	}

	// start - end points
	int have_start_end[] = {start_set, end_set};
	fwrite(have_start_end, sizeof *have_start_end, 2, fp);

	fwrite(&start_pos.x, sizeof start_pos.x, 1, fp);
	fwrite(&start_pos.y, sizeof start_pos.y, 1, fp);
	
	fwrite(&end_pos.x, sizeof end_pos.x, 1, fp);
	fwrite(&end_pos.y, sizeof end_pos.y, 1, fp);

	fclose(fp);
	return true;
}

bool Level::load(const char *fname)
{
	FILE *fp;

	if(!(fp = fopen(fname, "rb"))) {
		return false;
	}

	reset();

	int num_lines;
	fread(&num_lines, sizeof num_lines, 1, fp);

	for(int i=0; i<num_lines; i++) {
		int col;
		fread(&col, sizeof col, 1, fp);

		int collidable;
		fread(&collidable, sizeof col, 1, fp);

		int num_pt;
		fread(&num_pt, sizeof num_pt, 1, fp);

		PolyLine *pline = new PolyLine;
		pline->set_color(get_color(col));
		pline->set_width(line_width);

		for(int j=0; j<num_pt; j++) {
			Vector2 pt;

			fread(&pt.x, sizeof pt.x, 1, fp);
			fread(&pt.y, sizeof pt.y, 1, fp);

			pline->add_point(pt);
		}
		add_line(pline);
		fin_line(pline, collidable, col);
	}

	// start - end points
	int have_start_end[2];
	fread(have_start_end, sizeof *have_start_end, 2, fp);

	start_set = have_start_end[0];
	end_set = have_start_end[1];

	fread(&start_pos.x, sizeof start_pos.x, 1, fp);
	fread(&start_pos.y, sizeof start_pos.y, 1, fp);
	
	fread(&end_pos.x, sizeof end_pos.x, 1, fp);
	fread(&end_pos.y, sizeof end_pos.y, 1, fp);

	fclose(fp);
	return true;
}

void Level::set_space(SimSpace *sp)
{
	space = sp;
}

void Level::add_line(PolyLine *line)
{
	lines.push_back(line);
}

void Level::fin_line(PolyLine *line, bool colid, int col)
{
	line_colid.push_back(colid);
	line_color.push_back(col);

	if(colid) {
		add_line_to_space(line);
	}
}

void Level::add_line_to_space(PolyLine *line)
{
	if(!space) return;

	for(int i=0; i<line->get_point_count()-1; i++) {
		Vector2 v0 = line->get_point(i);
		Vector2 v1 = line->get_point(i + 1);

		Segment *segm = new Segment(static_body, v0, v1, line->get_width() / 2.0);
		segm->set_friction(0.6);
		segm->set_elasticity(1.0);
		space->add_static_shape(segm);

		this->segm.push_back(segm);
	}
}

void Level::set_char_pos(const Vector2 &pt)
{
	char_pos = pt;
}

bool Level::get_ground(const Vector2 &pt, Vector2 *gpt, Vector2 *norm) const
{
	float nearest_dist = FLT_MAX;
	Segment *nearest = 0;
	Vector2 nearest_pt;

	for(size_t i=0; i<segm.size(); i++) {
		Vector2 p1 = segm[i]->get_endpoint(0);
		Vector2 p2 = segm[i]->get_endpoint(1);

		if(p1.x > p2.x) {
			Vector2 tmp = p1;
			p1 = p2;
			p2 = tmp;
		}

		// if it's outside the range, skip it
		if(pt.x < p1.x || pt.x >= p2.x) {
			continue;
		}

		float t = (pt.x - p1.x) / (p2.x - p1.x);

		Vector2 proj_pt = lerp(p1, p2, t);
		float dist = pt.y - proj_pt.y;
		
		// skip all points above our point, or further away from
		// the nearest projection found so far.
		if(dist < 0.0 || dist > nearest_dist) {
			continue; 
		}

		nearest_dist = dist;
		nearest_pt = proj_pt;
		nearest = segm[i];
	}

	if(!nearest) {
		return false;
	}

	if(gpt)	*gpt = nearest_pt;
	if(norm) {
		Vector2 tang = nearest->get_endpoint(0) - nearest->get_endpoint(1);
		norm->x = -tang.y;
		norm->y = tang.x;

		if(norm->y < 0.0) {
			*norm = -*norm;
		}
	}
	return true;
}

bool Level::set_start(const Vector2 &pt)
{
	Vector2 tmp;
	if(get_ground(pt, &tmp, 0)) {
		start_pos = tmp + Vector2(0, 0.1);
		start_set = true;
		return true;
	}
	return false;
}

Vector2 Level::get_start() const
{
	return start_pos;
}

bool Level::set_end(const Vector2 &pt)
{
	Vector2 tmp;
	if(get_ground(pt, &tmp, 0)) {
		end_pos = tmp + Vector2(0, 0.1);
		end_set = true;
		return true;
	}
	return false;
}

Vector2 Level::get_end() const
{
	return end_pos;
}
/*
static void draw_point(const Vector2 &pt, float r, float g, float b)
{
	glDisable(GL_TEXTURE_2D);
	glPointSize(5.0);
	glBegin(GL_POINTS);
	glColor3f(r, g, b);
	glVertex2f(pt.x, pt.y);
	glEnd();
}*/

#define LINE_SAMPLES	16.0
void Level::draw() const
{
	for(size_t i=0; i<lines.size(); i++) {
		draw_chalk_line(lines[i], LINE_SAMPLES);
	}

	Vector2 pt;
	/*if(get_ground(char_pos, &pt, 0)) {
		draw_point(pt, 1, 0, 0);
	}*/
	if(start_set) {
		Vector2 arrow = start_pos + Vector2(0, 0.01);
		draw_chalk_line(arrow, start_pos + Vector2(0, 0.3), 0.05);
		draw_chalk_line(arrow - Vector2(0.025, 0.025), arrow + Vector2(0.1, 0.1), 0.05);
		draw_chalk_line(arrow - Vector2(-0.025, 0.025), arrow + Vector2(-0.1, 0.1), 0.05);
	}
	if(end_set) {
		Vector2 cent = end_pos + Vector2(0, 0.025);
		draw_chalk_line(cent - Vector2(0.1, 0.1), cent + Vector2(0.1, 0.1), 0.05, Color(0.85, 0.3, 0.2));
		draw_chalk_line(cent + Vector2(0.1, -0.1), cent + Vector2(-0.1, 0.1), 0.05, Color(0.85, 0.3, 0.2));
	}
}
