#include "opengl.h"
#include "polyline.h"

PolyLine::PolyLine()
{
	arc_length = -1.0;
	color = Color(1.0, 1.0, 1.0, 1.0);
	width = 0.1;
}

void PolyLine::set_point(int idx, const Vector2 &pt)
{
	points[idx] = pt;
}

const Vector2 &PolyLine::get_point(int idx) const
{
	return points[idx];
}

int PolyLine::get_point_count() const
{
	return points.size();
}

void PolyLine::set_width(float width)
{
	this->width = width;
}

float PolyLine::get_width() const
{
	return width;
}

void PolyLine::set_color(const Color &col)
{
	color = col;
}

const Color &PolyLine::get_color() const
{
	return color;
}

void PolyLine::add_point(const Vector2 &pt, float min_dist)
{
	if(!points.empty()) {
		if((pt - points.back()).length_sq() < SQ(min_dist)) {
			return;
		}
	}

	points.push_back(pt);
	arc_length = -1.0;
}

float PolyLine::get_length() const
{
	if(arc_length >= 0.0) {
		return arc_length;
	}

	arc_length = 0.0;
	for(size_t i=0; i<points.size() - 1; i++) {
		arc_length += (points[i + 1] - points[i]).length();
	}
	return arc_length;
}

Vector2 PolyLine::get_curve_point(float t) const
{
	if(points.empty()) {
		return Vector2(0, 0);
	} else if(points.size() == 1) {
		return points[0];
	}

	if(t < 0.0) {
		return points[0];
	} else if(t >= 1.0) {
		return points[points.size() - 1];
	}

	t *= (float)points.size() - 1;
	int seg = (int)t;
	t -= (float)floor(t);

	Vector2 cp[4] = {
		points[seg ? seg - 1 : seg],
		points[seg],
		points[seg + 1],
		points[seg >= (int)points.size() - 2 ? seg + 1 : seg + 2]
	};

	Vector2 res;
	res.x = catmull_rom_spline(cp[0].x, cp[1].x, cp[2].x, cp[3].x, t);
	res.y = catmull_rom_spline(cp[0].y, cp[1].y, cp[2].y, cp[3].y, t);
	return res;
}

Vector2 PolyLine::get_curve_normal(float t) const
{
	Vector2 dif = (get_curve_point(t + 0.01) - get_curve_point(t - 0.01)).normalized();
	return Vector2(-dif.y, dif.x);
}

#define TEX_X(x)	((x) / 1.33333 * 0.5 + 0.5)
#define TEX_Y(y)	(1.0 - (y) * 0.5 + 0.5)

void PolyLine::draw(float samples_per_unit) const
{
	float len = get_length();
	int samples = (int)(len * samples_per_unit);
	if(samples < 2) samples = 2;

	float rad = width / 2.0;

	glBegin(GL_QUADS);
	glColor3f(color.x, color.y, color.z);
	for(int i=0; i<samples-1; i++) {
		float t0 = (float)i / (float)(samples - 1);
		float t1 = (float)(i + 1) / (float)(samples - 1);

		Vector2 pt0 = get_curve_point(t0);
		Vector2 pt1 = get_curve_point(t1);

		Vector2 n0 = get_curve_normal(t0);
		Vector2 n1 = get_curve_normal(t1);

		//float s0 = i == 0 ? 0.0 : 0.5;
		//float s1 = i == samples - 2 ? 1.0 : 0.5;

		Vector2 v = pt0 - n0 * rad;
		//glTexCoord2f(s0, 0);
		//glMultiTexCoord2fARB(GL_TEXTURE1, TEX_X(v.x), TEX_Y(v.y));
		glTexCoord2f(TEX_X(v.x), TEX_Y(v.y));
		glVertex2f(v.x, v.y);

		v = pt1 - n1 * rad;
		//glTexCoord2f(s1, 0);
		//glMultiTexCoord2fARB(GL_TEXTURE1, TEX_X(v.x), TEX_Y(v.y));
		glTexCoord2f(TEX_X(v.x), TEX_Y(v.y));
		glVertex2f(v.x, v.y);

		v = pt1 + n1 * rad;
		//glTexCoord2f(s1, 1);
		//glMultiTexCoord2fARB(GL_TEXTURE1, TEX_X(v.x), TEX_Y(v.y));
		glTexCoord2f(TEX_X(v.x), TEX_Y(v.y));
		glVertex2f(v.x, v.y);

		v = pt0 + n0 * rad;
		//glTexCoord2f(s0, 1);
		//glMultiTexCoord2fARB(GL_TEXTURE1, TEX_X(v.x), TEX_Y(v.y));
		glTexCoord2f(TEX_X(v.x), TEX_Y(v.y));
		glVertex2f(v.x, v.y);
	}
	glEnd();
}
