#include <math.h>
#include "ragdoll.h"
#include "polyline.h"
#include "chalk.h"

static Box ragdoll_boxes[] = {
	Box(-20, 30, 20, -30),		// body
	Box(-8, 20, 8, -20),		// left upper leg
	Box(-8, 20, 8, -20),		// right upper leg
	Box(-8, 20, 8, -20),		// left lower leg
	Box(-8, 20, 8, -20),		// right lower leg
	Box(-20, 8, 20, -8),		// left upper arm
	Box(-20, 8, 20, -8),		// right upper arm
	Box(-20, 8, 20, -8),		// left lower arm
	Box(-20, 8, 20, -8),		// right lower arm
	Box(-15, 15, 15, -15)		// head
};
static int num_boxes = sizeof ragdoll_boxes / sizeof ragdoll_boxes[0];

static Vector2 bpos[] = {
	Vector2(0, 0),
	Vector2(-16, -50),
	Vector2(16, -50),
	Vector2(-16, -90),
	Vector2(16, -90),
	Vector2(-40, 27),
	Vector2(40, 27),
	Vector2(-80, 27),
	Vector2(80, 27),
	Vector2(0, 50)
};

static struct { int a, b; Vector2 pivot; float rmin, rmax; } links[] = {
	{0, 1, Vector2(-16, -30), -45, 20},
	{1, 3, Vector2(-16, -70), -10, 10},
	{0, 2, Vector2(16, -30), -20, 45},
	{2, 4, Vector2(16, -70), -10, 10},
	{5, 0, Vector2(-20, 27), -90, 75},
	{0, 6, Vector2(20, 27), -90, 75},
	{7, 5, Vector2(-60, 27), -90, 0},
	{6, 8, Vector2(60, 27), -90, 0},
	{9, 0, Vector2(0, 30), -30, 30}
};
static int num_links = sizeof links / sizeof links[0];


#define SCALE_FACTOR	0.0025

Box::Box() {}

Box::Box(float l, float t, float r, float b)
{
	Vector2 corners[] = {
		Vector2(r, t),
		Vector2(r, b),
		Vector2(l, b),
		Vector2(l, t)
	};

	for(int i=0; i<4; i++) {
		verts[i] = corners[i];
	}
}

Ragdoll::Ragdoll()
{	
	//cp_constraint_bias_coef = 0.9;

	for(int i=0; i<num_boxes; i++) {
		for(int j=0; j<4; j++) {
			ragdoll_boxes[i].verts[j] *= SCALE_FACTOR;
		}
		boxes.push_back(ragdoll_boxes[i]);

		const float mass = 50;
		Body *body = new Body(mass, calc_moment_polygon(mass, boxes[i].verts, 4));
		body->set_position(bpos[i] * SCALE_FACTOR);

		Shape *shape = new Poly(body, boxes[i].verts, 4);
		shape->set_elasticity(0.9);
		shape->set_friction(0.7);
		shape->set_group(1);

		bodies.push_back(body);
		shapes.push_back(shape);
	}

	for(int i=0; i<num_links; i++) {
		int a = links[i].a;
		int b = links[i].b;

		links[i].pivot *= SCALE_FACTOR;

		Constraint *joint = new PivotJoint(bodies[a], bodies[b], links[i].pivot);
		joints.push_back(joint);

		float rmin = DEG_TO_RAD(links[i].rmin);
		float rmax = DEG_TO_RAD(links[i].rmax);
		joint = new RotLimit(bodies[a], bodies[b], rmin, rmax);
		joints.push_back(joint);
	}
}


Ragdoll::~Ragdoll()
{
	for(size_t i=0; i<bodies.size(); i++) {
		delete bodies[i];
	}

	for(size_t i=0; i<shapes.size(); i++) {
		delete shapes[i];
	}

	for(size_t i=0; i<joints.size(); i++) {
		delete joints[i];
	}
}

void Ragdoll::set_color(const Color &col)
{
	color = col;
}

void Ragdoll::set_position(const Vector2 &pos)
{
	for(size_t i=0; i<bodies.size(); i++) {
		bodies[i]->set_position(bpos[i] * SCALE_FACTOR + pos);
		bodies[i]->set_angle(0.0);
	}
}

Vector2 Ragdoll::get_position() const
{
	return bodies[0]->get_position();
}

Body *Ragdoll::get_body(BodyPart part)
{
	return bodies[part];
}

Shape *Ragdoll::get_shape(BodyPart part)
{
	return shapes[part];
}

void Ragdoll::reset()
{
	for(size_t i=0; i<bodies.size(); i++) {
		bodies[i]->set_velocity(Vector2(0, 0));
	}
	set_position(Vector2(0, 0));
}

void Ragdoll::add_to_space(SimSpace *sp) const
{
	for(size_t i=0; i<bodies.size(); i++) {
		sp->add_body(bodies[i]);
	}

	for(size_t i=0; i<shapes.size(); i++) {
		sp->add_shape(shapes[i]);
	}

	for(size_t i=0; i<joints.size(); i++) {
		sp->add_constraint(joints[i]);
	}
}

void Ragdoll::remove_from_space(SimSpace *sp) const
{
	for(size_t i=0; i<bodies.size(); i++) {
		sp->remove_body(bodies[i]);
	}

	for(size_t i=0; i<shapes.size(); i++) {
		sp->remove_shape(shapes[i]);
	}

	for(size_t i=0; i<joints.size(); i++) {
		sp->remove_constraint(joints[i]);
	}
}

void Ragdoll::draw() const
{
	/*glBegin(GL_QUADS);
	glColor3f(1, 0, 0);
	for(size_t i=0; i<boxes.size(); i++) {
		for(int j=0; j<4; j++) {
			Vector2 v = boxes[i].verts[j].rotated(bodies[i]->get_angle());
			Vector2 p = bodies[i]->get_position();//.rotated(bodies[i]->get_angle());
			glVertex2f(p.x + v.x, p.y + v.y);
		}
	}
	glEnd();*/

	for(size_t i=0; i<boxes.size(); i++) {
		Box b;
		for(int j=0; j<4; j++) {
			Vector2 v = boxes[i].verts[j].rotated(bodies[i]->get_angle());
			Vector2 p = bodies[i]->get_position();
			b.verts[j] = p + v;
		}

		Vector2 pt[2];

		float length = (boxes[i].verts[0] - boxes[i].verts[3]).length();
		float width = (boxes[i].verts[0] - boxes[i].verts[1]).length();
		if(width > length) {
			float tmp = width;
			width = length;
			length = tmp;

			pt[0] = (b.verts[0] + b.verts[3]) / 2.0;
			pt[1] = (b.verts[1] + b.verts[2]) / 2.0;
		} else {
			pt[0] = (b.verts[0] + b.verts[1]) / 2.0;
			pt[1] = (b.verts[2] + b.verts[3]) / 2.0;
		}

		PolyLine pline;
		if(length / width < 2.0) {
			for(int j=0; j<5; j++) {
				pline.add_point((b.verts[j % 4] + b.verts[(j + 1) % 4]) / 2.0);
			}
			pline.set_width(width / 2.0);
		} else {
			pline.add_point(pt[0]);
			pline.add_point(pt[1]);
			pline.set_width(width);
		}

		pline.set_color(color);
		draw_chalk_line(&pline, 200.0);
	}
}

void Ragdoll::draw_outline() const
{/*
	glColor3f(0, 0, 1);
	for(size_t i=0; i<boxes.size(); i++) {
		glBegin(GL_LINE_LOOP);
		for(int j=0; j<4; j++) {
			Vector2 v = boxes[i].verts[j].rotated(bodies[i]->get_angle());
			Vector2 p = bodies[i]->get_position();//.rotated(bodies[i]->get_angle());
			glVertex2f(p.x + v.x, p.y + v.y);
		}
		glEnd();
	}*/
}
