#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "sdlglut.h"
#include "libfixgl/gl.h"

static int width, height;
static unsigned int disp_mode;

static void cleanup();
static void handle_event(SDL_Event *ev);
static void key_event(int key, bool pressed);
static void button_event(int bn, bool pressed);

static void (*disp_func)();
static void (*idle_func)();
static void (*reshape_func)(int, int);
static void (*keyb_func)(unsigned char, int, int);
static void (*keyb_up_func)(unsigned char, int, int);
static void (*spec_func)(int, int, int);
static void (*spec_up_func)(int, int, int);
static void (*mouse_func)(int, int, int, int);
static void (*motion_func)(int, int);
static void (*pasv_motion_func)(int, int);

SDL_Surface *fbsurf;
static int mouse_x, mouse_y;
static int cur_cursor;
static unsigned int bnpress;

#ifdef GP2X
static void handle_gp2x_joystick(SDL_Event *ev);
static void emulate_mouse(unsigned int dt);

static SDL_Joystick *joy;
static int gp2x_stick;
#endif

void glutInit(int *argcp, char **argv)
{
	SDL_Init(SDL_INIT_VIDEO | SDL_INIT_JOYSTICK | SDL_INIT_TIMER);
	SDL_ShowCursor(SDL_DISABLE);
#ifdef GP2X
	joy = SDL_JoystickOpen(0);
	SDL_JoystickEventState(SDL_ENABLE);
#endif

	atexit(cleanup);
}

void glutInitWindowSize(int w, int h)
{
	width = w;
	height = h;
}

void glutInitDisplayMode(unsigned int mode)
{
	disp_mode = mode;
}

void glutCreateWindow(const char *title)
{
	if(!(fbsurf = SDL_SetVideoMode(width, height, 16, SDL_SWSURFACE))) {
		fprintf(stderr, "set video mode failed\n");
		return;
	}
	SDL_WM_SetCaption(title, 0);
	SDL_ShowCursor(SDL_DISABLE);

	fglCreateContext();
	glViewport(0, 0, width, height);
}


void glutDisplayFunc(void (*func)(void))
{
	disp_func = func;
}

void glutIdleFunc(void (*func)(void))
{
	idle_func = func;
}

void glutReshapeFunc(void (*func)(int, int))
{
	reshape_func = func;
}

void glutKeyboardFunc(void (*func)(unsigned char, int, int))
{
	keyb_func = func;
}

void glutKeyboardUpFunc(void (*func)(unsigned char, int, int))
{
	keyb_up_func = func;
}

void glutSpecialFunc(void (*func)(int, int, int))
{
	spec_func = func;
}

void glutSpecialUpFunc(void (*func)(int, int, int))
{
	spec_up_func = func;
}

void glutMouseFunc(void (*func)(int, int, int, int))
{
	mouse_func = func;
}

void glutMotionFunc(void (*func)(int, int))
{
	motion_func = func;
}

void glutPassiveMotionFunc(void (*func)(int, int))
{
	pasv_motion_func = func;
}


void glutMainLoop(void)
{
	if(reshape_func) {
		reshape_func(width, height);
	}

	if(disp_func) {
		disp_func();
	}

	for(;;) {
		SDL_Event ev;

		if(idle_func) {
			while(SDL_PollEvent(&ev)) {
				handle_event(&ev);
			}
#ifdef GP2X
			static unsigned int prev;
			unsigned int msec = SDL_GetTicks();;
			emulate_mouse(msec - prev);
			prev = msec;
#endif
			if(idle_func) idle_func();

		} else {
			SDL_WaitEvent(&ev);
			handle_event(&ev);
		}
	}
}

static void draw_cursor(uint16_t *fb, int x, int y)
{
	fb += y * width + x;

	*fb = 0xffff;

	if(x > 1) *(fb - 1) = 0xffff;
	if(x > 2) *(fb - 2) = 0xffff;

	if(x < width - 1) *(fb + 1) = 0xffff;
	if(x < width - 2) *(fb + 2) = 0xffff;

	if(y > 1) *(fb - width) = 0xffff;
	if(y > 2) *(fb - (width << 1)) = 0xffff;

	if(y < height - 1) *(fb + width) = 0xffff;
	if(y < height - 2) *(fb + (width << 1)) = 0xffff;
}

void glutSwapBuffers(void)
{
	uint16_t *pixels = (uint16_t*)fglGetFrameBuffer();

	if(cur_cursor != GLUT_CURSOR_NONE) {
		draw_cursor(pixels, mouse_x, mouse_y);
	}

	if(SDL_MUSTLOCK(fbsurf)) {
		SDL_LockSurface(fbsurf);
		memcpy(fbsurf->pixels, pixels, width * height * sizeof *pixels);
		SDL_UnlockSurface(fbsurf);
	} else {
		memcpy(fbsurf->pixels, pixels, width * height * sizeof *pixels);
	}

	SDL_Flip(fbsurf);
}

void glutPostRedisplay(void)
{
	SDL_Event ev;
	ev.type = SDL_VIDEOEXPOSE;

	SDL_PushEvent(&ev);
}

void glutSetCursor(int cursor)
{
	cur_cursor = cursor;
}

int glutGet(unsigned int what)
{
	switch(what) {
	case GLUT_WINDOW_X:
	case GLUT_WINDOW_Y:
		return 0;

	case GLUT_WINDOW_WIDTH:
		return width;

	case GLUT_WINDOW_HEIGHT:
		return height;

	case GLUT_ELAPSED_TIME:
		return SDL_GetTicks();
	}

	return -1;
}

static void cleanup()
{
	fglDestroyContext();
	SDL_Quit();
}

static void handle_event(SDL_Event *ev)
{

	switch(ev->type) {
	case SDL_KEYDOWN:
		key_event(ev->key.keysym.sym, true);
		break;

	case SDL_KEYUP:
		key_event(ev->key.keysym.sym, false);
		break;

	case SDL_VIDEOEXPOSE:
		if(disp_func) {
			disp_func();
		}
		break;

	case SDL_VIDEORESIZE:
		width = ev->resize.w;
		height = ev->resize.h;

		if(reshape_func) {
			reshape_func(width, height);
		}
		break;

	case SDL_MOUSEMOTION:
		mouse_x = ev->motion.x;
		mouse_y = ev->motion.y;

		if(bnpress) {
			if(motion_func) motion_func(mouse_x, mouse_y);
		} else {
			if(pasv_motion_func) pasv_motion_func(mouse_x, mouse_y);
		}
		break;

	case SDL_MOUSEBUTTONDOWN:
		button_event(ev->button.button, true);
		break;

	case SDL_MOUSEBUTTONUP:
		button_event(ev->button.button, false);
		break;

#ifdef GP2X
	case SDL_JOYBUTTONDOWN:
	case SDL_JOYBUTTONUP:
		handle_gp2x_joystick(ev);
		break;
#endif

	case SDL_QUIT:
		exit(0);
	}
}


static void key_event(int key, bool pressed)
{
	if(pressed) {
		if(key > 0xff) {
			if(spec_func) spec_func(key, mouse_x, mouse_y);
		} else {
			if(keyb_func) keyb_func(key, mouse_x, mouse_y);
		}
	} else {
		if(key > 0xff) {
			if(spec_up_func) spec_up_func(key, mouse_x, mouse_y);
		} else {
			if(keyb_up_func) keyb_up_func(key, mouse_x, mouse_y);
		}
	}
}

static void button_event(int bn, bool pressed)
{
	if(pressed) {
		bnpress |= (1 << (bn - 1));
		if(mouse_func) {
			mouse_func(bn - 1, GLUT_DOWN, mouse_x, mouse_y);
		}
	} else {
		bnpress &= ~(1 << (bn - 1));
		if(mouse_func) {
			mouse_func(bn - 1, GLUT_UP, mouse_x, mouse_y);
		}
	}
}

#ifdef GP2X
static int press_count;

static void handle_gp2x_joystick(SDL_Event *ev)
{
	assert(ev->type == SDL_JOYBUTTONDOWN || ev->type == SDL_JOYBUTTONUP);

	int bn = ev->jbutton.button;
	bool down = ev->jbutton.state == SDL_PRESSED;

	if(bn >= 0 && bn < 8) {
		if(down) {
			gp2x_stick = bn;
			press_count++;
		} else {
			press_count--;
		}
		if(!press_count) {
			gp2x_stick = -1;
		}
	}

	switch(bn) {
	case GP2X_VK_UP:
		key_event(GLUT_KEY_UP, down);
		break;

	case GP2X_VK_DOWN:
		key_event(GLUT_KEY_DOWN, down);
		break;

	case GP2X_VK_LEFT:
		key_event(GLUT_KEY_LEFT, down);
		break;

	case GP2X_VK_RIGHT:
		key_event(GLUT_KEY_RIGHT, down);
		break;

	case GP2X_VK_FR:
		button_event(SDL_BUTTON_LEFT, down);
		break;

	case GP2X_VK_FL:
		button_event(SDL_BUTTON_RIGHT, down);
		key_event('c', down);
		break;

	case GP2X_VK_STICK_PUSH:
		key_event('`', down);
		break;

	case GP2X_VK_VOL_DOWN:
		key_event(GLUT_KEY_HOME, down);
		break;

	case GP2X_VK_VOL_UP:
		key_event(GLUT_KEY_END, down);
		break;

	case GP2X_VK_FA:
		key_event('\t', down);
		break;

	case GP2X_VK_FB:
		key_event('\n', down);
		break;

	case GP2X_VK_FY:
		button_event(SDL_BUTTON_MIDDLE, down);
		break;

	case GP2X_VK_FX:
		key_event(27, down);
		break;

	case GP2X_VK_START:
		key_event('l', down);
		break;

	case GP2X_VK_SELECT:
		key_event('s', down);
		break;
	}
}

enum {
	STICK_U,
	STICK_UL,
	STICK_L,
	STICK_DL,
	STICK_D,
	STICK_DR,
	STICK_R,
	STICK_UR
};

static void emulate_mouse(unsigned int dt)
{
	if(!press_count) return;

	int offs = dt >> 4;
	if(!offs) offs = 1;

	bool mouse_moved = false;

	if(gp2x_stick == STICK_R || gp2x_stick == STICK_UR || gp2x_stick == STICK_DR) {
		if(mouse_x < width - 1) {
			mouse_x += offs;
			mouse_moved = true;
		}
	}
	if(gp2x_stick == STICK_L || gp2x_stick == STICK_UL || gp2x_stick == STICK_DL) {
		if(mouse_x > 0) {
			mouse_x -= offs;
			mouse_moved = true;
		}
	}
	if(gp2x_stick == STICK_D || gp2x_stick == STICK_DL || gp2x_stick == STICK_DR) {
		if(mouse_y < width - 1) {
			mouse_y += offs;
			mouse_moved = true;
		}
	}
	if(gp2x_stick == STICK_U || gp2x_stick == STICK_UL || gp2x_stick == STICK_UR) {
		if(mouse_y > 0) {
			mouse_y -= offs;
			mouse_moved = true;
		}
	}

	if(mouse_moved) {
		if(bnpress) {
			if(motion_func) motion_func(mouse_x, mouse_y);
		} else {
			if(pasv_motion_func) pasv_motion_func(mouse_x, mouse_y);
		}
	}
}
#endif	// GP2X
